"""
Logic related to exporting a trained model with a corresponding
preprocessing pipeline.
"""
import abc
import datetime
import logging
from copy import deepcopy
from pathlib import Path
from typing import Optional

import joblib
from sklearn.pipeline import Pipeline

from . import _internals

logger = logging.getLogger(f"stream.{__name__}")


def export_model(setup: _internals.Setup, model, target_dir: str) -> None:
    """Exports provided `model` with preprocessing pipeline stored in
    setup.

    The output file is binarized using joblib and saved to `target_dir`.
    The output filename follows the syntax
      `model_{date}_{algo}_#{number}.joblib`
    where
    - `date` is the current date in format YYYY-MM-DD
    - `algo` is the utilised model algorithm, and
    - `number` is a number assigned when other models with the same
    name already exist.

    Example
    -------
    >>> pt.modelling.export_model(
            setup=setup,
            model=model_list[0],
            target_dir="./",
        )

    Parameters
    ----------
    setup : _internals.Setup
        Setup object containing preprocessing pipeline.

    model :
        Trained model instance.

    target_dir : str
        Target directory for saving the output file to.
    """
    full_pipe = _combine_pipeline_and_model(prep_pipe=setup.prep_pipe, model=model)
    export_name = _get_export_name(target_dir=target_dir, model=model)
    joblib.dump(full_pipe, Path(target_dir) / export_name)
    logger.info(f"Exported modelling pipeline to '{Path(target_dir) / export_name}'")


def _combine_pipeline_and_model(prep_pipe: Pipeline, model) -> Pipeline:
    """Adds `model` as last step to `prep_pipe`."""
    # Making deep copy, so that the model-step is not added to the
    # original pipeline object.
    prep_pipe_copy = deepcopy(prep_pipe)
    prep_pipe_copy.steps.append(("model", model))
    return prep_pipe_copy


def _get_export_name(target_dir: str, model) -> str:
    """Generates filename for `full_pipe`."""
    date = str(datetime.date.today())
    algo = _get_algo_abbreviation(model_type=type(model))
    temp_name = f"model_{date}_{algo}_#"
    new_number = _get_new_number(target_dir=target_dir, temp_name=temp_name)
    return (
        f"model_{date}_{algo}_#{new_number}.joblib"
        if new_number
        else f"model_{date}_{algo}_#1.joblib"
    )


def _get_algo_abbreviation(model_type: abc.ABCMeta) -> str:
    """Returns string abbreviation for provided algorithm."""
    algo_dict = _internals.get_algo_dict()
    return [key for key in algo_dict.keys() if algo_dict[key] == model_type][0]


def _get_new_number(target_dir: str, temp_name: str) -> Optional[int]:
    """Determines the number value for the new file name by
    checking `target_dir` for pre-existing files with matching
    name pattern.

    Parameters
    ----------
    target_dir : str
        Target directory for new file. This is also where for
        pre-existing files of similar name is checked.

    temp_name : str
        Basic name structure for new file.

    Returns
    -------
    Optional[int]
        Integer value of the new number, or None.
        None is returned when no pre-existing file is found.
    """
    files_in_target_dir = [
        str(path.name) for path in Path(target_dir).glob("./*") if path.is_file()
    ]
    files_matching_pattern = [
        filename for filename in files_in_target_dir if temp_name in filename
    ]
    if len(files_matching_pattern) > 0:
        last_entry = sorted(files_matching_pattern)[-1]
        last_number = int(last_entry.rsplit("#")[-1].split(".joblib")[0])
        return last_number + 1
    else:
        return None



import logging

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder

logger = logging.getLogger(f"stream.{__name__}")


def get_prep_pipeline() -> Pipeline:
    """Return unfitted preprocessing pipeline.

    Preprocessing steps include:
      1) One-Hot-Encoding of 'object' columns

    Returns
    -------
    Pipeline
        Unfitted preprocessing pipeline.
    """
    prep_pipe = Pipeline(steps=[("ohe", OHE())])
    step_names = [step[0] for step in prep_pipe.steps]
    logger.info(f"Created preprocessing pipeline with following steps: {step_names}")
    return prep_pipe


class OHE(BaseEstimator, TransformerMixin):
    """Transformer class performing One-Hot-Encoding of 'object' typed columns."""

    def __init__(self):
        self.ohe = OneHotEncoder(handle_unknown="ignore")
        self.col_names = None

    def fit(self, X, y=None):
        categorical_data = X.select_dtypes(include=("object"))
        if len(categorical_data.columns) > 0:
            self.ohe.fit(categorical_data)
            self.col_names = self.ohe.get_feature_names_out(categorical_data.columns)
        return self

    def transform(self, X, y=None):
        # Checking whether encoder was fitted
        if self.col_names is not None:
            non_categorical_data = X.select_dtypes(exclude=("object"))
            categorical_data = X.select_dtypes(include=("object"))
            ohe_data = self.ohe.transform(categorical_data).toarray()
            ohe_df = pd.DataFrame(
                data=ohe_data, columns=self.col_names, index=non_categorical_data.index
            ).astype("uint8")
            return pd.concat([non_categorical_data, ohe_df], axis=1)
        else:
            return X
